/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 * 
 * Copyright (©) 2005,2006,2007,2008 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2002,2007 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992,2007 Frank Tore Johansen
 * 
 * Deliantra is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

//
// cfperl.h perl interface
//
#ifndef CFPERL_H__
#define CFPERL_H__

#include <cstdarg>
#include <cstdio>
#include <bitset>

using namespace std;

#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>

#include <EVAPI.h>
#include <CoroAPI.h>

#include "util.h"
#include "keyword.h"
#include "dynbuf.h"
#include "callback.h"

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// optimisations/workaround for functions requiring my_perl in scope (anti-bloat)
#undef localtime
#undef srand48
#undef drand48
#undef srandom
#undef opendir
#undef readdir
#undef closedir
#undef getprotobyname
#undef gethostbyname
#undef ctime
#undef strerror
#undef _

// same here, massive symbol spamming
#undef do_open
#undef do_close
#undef ref
#undef seed

// perl bug #40256: perl does overwrite those with reentrant versions
// but does not initialise their state structures.
#undef random
#undef crypt

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#define _(msg)  (msg)
#define N_(msg) (msg)

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// some macros to simplify perl in-calls

#define CHECK_ERROR			\
  if (SvTRUE (ERRSV))			\
    LOG (llevError, "runtime error in %s: %s", __func__, SvPVutf8_nolen (ERRSV));

#define CALL_BEGIN(args) dSP; ENTER; SAVETMPS; PUSHMARK (SP); EXTEND (SP, args)
#define CALL_ARG_SV(sv) PUSHs (sv_2mortal (sv)) // separate because no refcount inc
#define CALL_ARG(expr) PUSHs (sv_2mortal (to_sv (expr)))
#define CALL_CALL(name, flags) PUTBACK; int count = call_pv (name, (flags) | G_EVAL); SPAGAIN;
#define CALL_END CHECK_ERROR; FREETMPS; LEAVE

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void cfperl_init ();
void cfperl_main ();
void cfperl_tick ();
void cfperl_emergency_save ();
void cfperl_cleanup (int make_core);
void cfperl_make_book (object *book, int level);
void cfperl_expand_cfpod (player *pl, std::string &msg);
void cfperl_send_msg (client *ns, int color, const char *type, const char *msg);
int cfperl_can_merge (object *ob1, object *ob2);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#if IVSIZE >= 8
  typedef IV val64;
  typedef IV uval64;
# define newSVval64 newSViv
# define SvVAL64 SvIV
# define newSVuval64 newSVuv
# define SvUVAL64 SvUV
#else
  typedef double val64;
  typedef double uval64;
# define newSVval64 newSVnv
# define SvVAL64 SvNV
# define newSVuval64 newSVnv
# define SvUVAL64 SvNV
#endif

extern tstamp runtime; // virtual server time, excluding time jumps and lag
extern tstamp NOW;     // real time of current server tick

enum event_klass
{
  KLASS_NONE,
  KLASS_GLOBAL,
  KLASS_ATTACHABLE,
  KLASS_CLIENT,
  KLASS_PLAYER,
  KLASS_OBJECT,
  KLASS_MAP,
  KLASS_COMMAND,
};

enum event_type
{
# define def(klass,name) EVENT_ ## klass ## _ ## name,
#  include "eventinc.h"
# undef def
  NUM_EVENT_TYPES
};

// in which global events or per-type events are we interested
extern bitset<NUM_EVENT_TYPES> ev_want_event;
extern bitset<NUM_TYPES>       ev_want_type;

#define ARG_AV(o)         DT_AV    , static_cast<AV *>   (o)
#define ARG_INT(v)        DT_INT   , static_cast<int>    (v)
#define ARG_INT64(v)      DT_INT64 , static_cast<sint64> (v)
#define ARG_DOUBLE(v)     DT_DOUBLE, static_cast<double> (v)
#define ARG_STRING(v)     DT_STRING, static_cast<const char *> (v)
#define ARG_DATA(s,l)     DT_DATA  , static_cast<const void *> (s), int (l)
#define ARG_OBJECT(o)     DT_OBJECT, (void *)(static_cast<object *>    (o))
#define ARG_MAP(o)        DT_MAP   , (void *)(static_cast<maptile *>   (o))
#define ARG_PLAYER(o)     DT_PLAYER, (void *)(static_cast<player *>    (o))
#define ARG_ARCH(o)       DT_ARCH  , (void *)(static_cast<archetype *> (o))
#define ARG_CLIENT(o)     DT_CLIENT, (void *)(static_cast<client *>    (o))
#define ARG_PARTY(o)      DT_PARTY , (void *)(static_cast<party *>     (o))
#define ARG_REGION(o)     DT_REGION, (void *)(static_cast<region *>    (o))

// the ", ## __VA_ARGS" is, unfortunately, a gnu-cpp extension
#define INVOKE(obj,event, ...) (expect_false ((obj)->should_invoke (event)) ? (obj)->invoke (event, ## __VA_ARGS__, DT_END) : 0)
#define INVOKE_GLOBAL(event, ...)          INVOKE (&gbl_ev, EVENT_ ## GLOBAL     ## _ ## event, ## __VA_ARGS__)
#define INVOKE_ATTACHABLE(event, obj, ...) INVOKE (obj    , EVENT_ ## ATTACHABLE ## _ ## event, ## __VA_ARGS__)
#define INVOKE_OBJECT(event, obj, ...)     INVOKE (obj    , EVENT_ ## OBJECT     ## _ ## event, ## __VA_ARGS__)
#define INVOKE_CLIENT(event, obj, ...)     INVOKE (obj    , EVENT_ ## CLIENT     ## _ ## event, ## __VA_ARGS__)
#define INVOKE_PLAYER(event, obj, ...)     INVOKE (obj    , EVENT_ ## PLAYER     ## _ ## event, ## __VA_ARGS__)
#define INVOKE_MAP(event, obj, ...)        INVOKE (obj    , EVENT_ ## MAP        ## _ ## event, ## __VA_ARGS__)

//TODO should index into @result
#define RESULT(idx,type) cfperl_result_ ## type (idx)
#define RESULT_DOUBLE(idx) RESULT(idx, DOUBLE)
#define RESULT_INT(idx)    RESULT(idx, INT)

double cfperl_result_DOUBLE (int idx);
int cfperl_result_INT (int idx);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

INTERFACE_CLASS (attachable)
struct attachable : refcnt_base
{
  static MGVTBL vtbl;

  static unordered_vector<attachable *> mortals;
  MTH static void check_mortals ();

  enum {
    F_DESTROYED   = 0x01,
    F_DEBUG_TRACE = 0x02,
  };
  int ACC (RW, flags);

  // object is delete'd after the refcount reaches 0
  MTH int refcnt_cnt () const;
  // check wether the object has died and destroy
  MTH void refcnt_chk () { if (expect_false (refcnt <= 0)) do_check (); }

  // destroy the object unless it was already destroyed
  // this politely asks everybody interested the reduce
  // the refcount to 0 as soon as possible.
  MTH void destroy ();

  // return wether an object was destroyed already
  MTH bool destroyed () const { return flags & F_DESTROYED; }

  virtual void gather_callbacks (AV *&callbacks, event_type event) const;

#if 0
private:
  static refcounted *rc_first;
  refcounted *rc_next;
#endif

  HV *self;     // CF+ perl self
  AV *cb;       // CF+ callbacks
  shstr attach; // generic extension attachment information

  void sever_self (); // sever this object from its self, if it has one.
  void optimise ();   // possibly save some memory by destroying unneeded data

  attachable ()
  : flags (0), self (0), cb (0), attach (0)
  {
  }

  attachable (const attachable &src)
  : flags (0), self (0), cb (0), attach (src.attach)
  {
  }

  // set a custom key to the given value, or delete it if value = 0
  void set_key (const char *key, const char *value = 0, bool is_utf8 = 0);

  void set_key_text (const char *key, const char *value = 0)
  {
    set_key (key, value, 1);
  }

  void set_key_data (const char *key, const char *value = 0)
  {
    set_key (key, value, 0);
  }

  attachable &operator =(const attachable &src);

  // used to _quickly_ device wether to shortcut the evaluation
  bool should_invoke (event_type event)
  {
    return ev_want_event [event] || cb;
  }

  bool invoke (event_type event, ...);

  MTH void instantiate ();
  void reattach ();

protected:
  // do the real refcount checking work
  void do_check ();

  // the method that does the real destroy work
  virtual void do_destroy ();

  // destructor is protected, should not be called by anybody
  virtual ~attachable ();
};

// the global object is a pseudo object that cares for the global events
struct global : attachable
{
  void gather_callbacks (AV *&callbacks, event_type event) const;
};

extern struct global gbl_ev;

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

INTERFACE_CLASS(object_freezer)
struct object_freezer : dynbuf_text
{
  AV *av;

  object_freezer ();
  ~object_freezer ();

  void put (attachable *ext);

  // used only for user-defined key-value pairs
  void put (const shstr &k, const shstr &v)
  {
    add (k);

    if (expect_true (v))
      add (' '), add (v);

    add ('\n');
  }

  template<typename T>
  void put_kw_string (keyword k, const T &v)
  {
    int klen = keyword_len [k];
    int vlen = v ? strlen (v) + 1 : 0;

    char *p = (char *)alloc (klen + vlen + 1);

    memcpy (p, keyword_str [k], klen); p += klen;

    if (expect_true (v))
      {
        *p++ = ' ';  vlen--;
        memcpy (p, v, vlen); p += vlen;
      }

    *p = '\n';
  }

  void put (keyword k, const char *v = 0)
  {
    put_kw_string (k, v);
  }

  void put (keyword k, const shstr &v)
  {
    put_kw_string (k, v);
  }

  void put (keyword k, double v)
  {
    force (MAX_KEYWORD_LEN + 2 + 32);
    fadd (keyword_str [k], keyword_len [k]);
    fadd (' ');
    falloc (sprintf (ptr, "%.7g", v));
    fadd ('\n');
  }

  void put_(keyword k, sint64 v)
  {
    force (MAX_KEYWORD_LEN + 2 + sint64_digits);
    fadd (keyword_str [k], keyword_len [k]);
    fadd (' ');
    add (v);
    fadd ('\n');
  }

  void put_(keyword k, sint32 v)
  {
    force (MAX_KEYWORD_LEN + 2 + sint32_digits);
    fadd (keyword_str [k], keyword_len [k]);
    fadd (' ');
    add (v);
    fadd ('\n');
  }

  void put (keyword k,          float     v) { put (k, (double)v); }
  void put (keyword k,   signed char      v) { put_(k, (sint32)v); }
  void put (keyword k, unsigned char      v) { put_(k, (sint32)v); }
  void put (keyword k,   signed short     v) { put_(k, (sint32)v); }
  void put (keyword k, unsigned short     v) { put_(k, (sint32)v); }
  void put (keyword k,   signed int       v) { put_(k, (sint32)v); }
  void put (keyword k, unsigned int       v) { put_(k, (sint64)v); }
  void put (keyword k,   signed long      v) { put_(k, (sint64)v); }
  void put (keyword k, unsigned long      v) { put_(k, (sint64)v); }
  void put (keyword k,   signed long long v) { put_(k, (sint64)v); }
  void put (keyword k, unsigned long long v) { put_(k, (sint64)v); }

  void put (keyword kbeg, keyword kend, const shstr &v)
  {
    force (MAX_KEYWORD_LEN + 1);
    fadd (keyword_str [kbeg], keyword_len [kbeg]); fadd ('\n');

    if (expect_true (v))
      {
        add (v);
        add ('\n');
      }

    force (MAX_KEYWORD_LEN + 1);
    fadd (keyword_str [kend], keyword_len [kend]); fadd ('\n');
  }

  void put (keyword k, archetype *v);
  void put (keyword k, treasurelist *v);
  void put (keyword k, faceinfo *v);

  template<typename T>
  void put (keyword k, const refptr<T> &v)
  {
    put (k, (T *)v);
  }

  MTH bool save (const_octet_string path);
  char *as_string (); // like strdup

  operator bool () { return !!av; }
};

INTERFACE_CLASS(object_thawer)
struct object_thawer
{
  char *line; // current beginning of line
  SV *text; // text part
  AV *av;   // perl part
  int linenum;
  keyword kw;
  char *kw_str; // the keyword parsed, as string
  char *value;  // the value, or 0 if no value
  const char *value_nn; // the value, or the empty string if no value
  const char *name;

  operator bool () const { return !!text; }

  object_thawer (const char *path = 0);
  object_thawer (const char *data, AV *perlav);
  ~object_thawer ();

  void get (attachable *obj, int oid);

  // parse next line
  MTH void next ();
  // skip the current key-value (usually fetch next line, for
  // multiline-fields, skips till the corresponding end-kw
  MTH void skip ();
  MTH void skip_block (); // skips till and over KW_end

  char *get_str () { return value; } // may be 0
  void get_ml (keyword kend, shstr &sh);

  void get_ornull (shstr &sh) const { sh = value; }
  void get (shstr &sh)        const { sh = value; } // might want to check for non-null here

  bool   get_bool   () const { return *value_nn == '1'; }
  sint32 get_sint32 () const;
  sint64 get_sint64 () const { return strtoll (value_nn, 0, 10); }
  double get_double () const { return strtod  (value_nn, 0); }

  void get (float  &v) { v = get_double (); }
  void get (double &v) { v = get_double (); }

  void get (bool   &i) { i = get_bool   (); }
  void get (sint8  &i) { i = get_sint32 (); }
  void get (uint8  &i) { i = get_sint32 (); }
  void get (sint16 &i) { i = get_sint32 (); }
  void get (uint16 &i) { i = get_sint32 (); }
  void get (sint32 &i) { i = get_sint32 (); }

  void get (uint32 &i) { i = get_sint64 (); }
  void get (sint64 &i) { i = get_sint64 (); }

  MTH void parse_warn (const char *msg);
  MTH bool parse_error (const char *type = 0, const char *name = 0, bool skip = true);

  struct delayed_ref {
    attachable *op;
    object_ptr *ptr;
    const char *ref;
  };
  std::vector<delayed_ref> delrefs;

  void delayed_deref (attachable *op, object_ptr &ptr, const char *ref);
  MTH void resolve_delayed_derefs (bool deref = true);
};

//TODO: remove
char *fgets (char *s, int n, object_thawer &thawer);

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct coroapi {
  static struct CoroAPI *GCoroAPI;

  static EV_ATOMIC_T cede_pending;

  static int nready () {                   return CORO_NREADY; }
  static int cede   () { cede_pending = 0; return CORO_CEDE  ; }

  static void do_cede_to_tick ();

  // actually cede's far more often
  static bool cede_to_tick ()
  {
    if (expect_true (!cede_pending))
      return false;

    do_cede_to_tick ();
    return true;
  }

  static void wait_for_tick ();
  static void wait_for_tick_begin ();
};

struct evapi
{
  static struct EVAPI *GEVAPI;
};

struct iow : ev_io, evapi, callback<void (iow &, int)>
{
  static void thunk (EV_P_ struct ev_io *w_, int revents)
  {
    iow &w = *static_cast<iow *>(w_);
    
    w (w, revents);
  }

  template<class O, class M>
  iow (O object, M method)
  : callback<void (iow &, int)> (object, method)
  {
    ev_init ((ev_io *)this, thunk);
  }

  void prio (int prio)
  {
    ev_set_priority ((ev_io *)this, prio);
  }

  void set (int fd, int events)
  {
    ev_io_set ((ev_io *)this, fd, events);
  }

  int poll () const { return events; }

  void poll (int events);

  void start ()
  {
    ev_io_start (EV_DEFAULT, (ev_io *)this);
  }

  void stop ()
  {
    ev_io_stop (EV_DEFAULT, (ev_io *)this);
  }

  ~iow ()
  {
    stop ();
  }
};

struct asyncw : ev_async, evapi, callback<void (ev_async &, int)>
{
  static void thunk (EV_P_ struct ev_async *w_, int revents)
  {
    asyncw &w = *static_cast<asyncw *>(w_);
    
    w (w, revents);
  }

  template<class O, class M>
  asyncw (O object, M method)
  : callback<void (asyncw &, int)> (object, method)
  {
    ev_init ((ev_async *)this, thunk);
  }

  void start ()
  {
    ev_async_start (EV_DEFAULT, (ev_async *)this);
  }

  void stop ()
  {
    ev_async_stop (EV_DEFAULT, (ev_async *)this);
  }

  ~asyncw ()
  {
    stop ();
  }
};

#endif

